from typing import Literal

import mwapi  # type: ignore

from toolforge_i18n._user_agent import get_user_agent


def get_gender_by_user_name(user_name: str | None) -> Literal['m', 'f', 'n']:
    """Get the gender of a named user on Wikimedia sites.

    This gets the gender from Meta-Wiki –
    hopefully the user set it as a global preference,
    not just on one other wiki.
    None may be used to represent an unknown user (e.g. not logged in),
    who will be treated as having neuter gender.
    """
    if user_name is None:
        return 'n'
    session = mwapi.Session('https://meta.wikimedia.org', user_agent=get_user_agent())
    response = session.get(action='query', list=['users'], usprop=['gender'], ususers=[user_name], formatversion=2)
    mapping: dict[str, Literal['m', 'f', 'n']] = {
        'male': 'm',
        'female': 'f',
        'unknown': 'n',
    }
    return mapping[response['query']['users'][0]['gender']]



from collections.abc import Callable
from typing import cast

import flask
import werkzeug
from markupsafe import Markup

from toolforge_i18n._formatters import I18nFormatter
from toolforge_i18n._language_info import lang_bcp47_to_mw, lang_dir, lang_fallbacks, lang_mw_to_bcp47
from toolforge_i18n._translations import load_translations


def init_html_language_codes() -> None:
    """Initialize the stack of HTML language codes.

    The stack is used to track whether a message needs to be wrapped
    in a new HTML element to set the lang= and dir= attributes,
    or whether it’s in the same language as the surrounding markup.
    It’s maintained by message(), push_html_lang() and pop_html_lang().
    """
    flask.g.html_language_codes = []


def push_html_lang(language_code: str) -> Markup:
    """Push an HTML language code to the stack.

    Many tools will not need to call this,
    as it’s called by the message() function automatically.
    However, if you also add localized text from other sources than messages,
    you should call this function with the MediaWiki language code you are using;
    for example, in a Jinja2 template:

        <span {{ push_html_lang(label.language) }}>
          {{ label.value }}
        </span{{ pop_html_lang(label.language) }}>
    """
    html_language_code = lang_mw_to_bcp47(language_code)
    flask.g.html_language_codes.append(html_language_code)
    return Markup(r'lang="{}" dir="{}"').format(html_language_code, lang_dir(html_language_code))


def add_lang_if_needed(message: Markup, language_code: str) -> Markup:
    """Wrap the given message in a language-tagged element if necessary.

    Given a (formatted) message in a certain language (MediaWiki language code),
    wrap it in a <span> with lang= and dir= attributes
    if the current language on top of the stack is different.
    Note that message() calls this function automatically,
    so you generally don’t need to use this function yourself.
    """
    if flask.g.html_language_codes and flask.g.html_language_codes[-1] == language_code:
        return message
    return Markup('<span {}>{}</span{}>').format(push_html_lang(language_code), message, pop_html_lang(language_code))


def pop_html_lang(language_code: str) -> Markup:
    """Pop an HTML language code from the stack.

    See push_html_lang() for details.
    """
    html_language_code = lang_mw_to_bcp47(language_code)
    assert flask.g.html_language_codes.pop() == html_language_code
    return Markup(r'')


def assert_html_language_codes_empty(response: werkzeug.Response) -> werkzeug.Response:
    """Assert that the stack of HTML language codes is depleted.

    This is called at the end of a request;
    if the assertion fails, some push_html_lang() call is missing
    a corresponding pop_html_lang() call.
    """
    assert flask.g.html_language_codes == []
    return response


def interface_language_code_from_request(translations: dict[str, dict[str, str]]) -> str:
    """Default implementation to determine the language code of a request.

    This function supports the ?uselang= URL parameter
    and otherwise determines the language based on the request’s Accept-Language header.
    You may want to override this method to implement a persistent language preference;
    to keep the features mentioned above,
    your implementation should generally look like this:

        from toolforge_i18n import interface_language_code_from_request

        def interface_language_code(translations):
            # ?uselang= takes precedence if present
            if 'uselang' in flask.request.args:
                return interface_language_code_from_request(translations)
            # try persistent language preference (e.g. from flask.session) next
            # ...
            # finally, fall back to Accept-Language:
            return interface_language_code_from_request(translations)

        # ...later, pass the implementation into ToolforgeI18n:
        i18n = ToolforgeI18n(app, interface_language_code)
    """
    if 'uselang' in flask.request.args:
        return flask.request.args['uselang']
    available_bcp47_languages = [lang_mw_to_bcp47(code) for code in translations]
    best_bcp47_language = flask.request.accept_languages.best_match(available_bcp47_languages, 'en')
    return lang_bcp47_to_mw(best_bcp47_language)


def _message_with_language(message_code: str) -> tuple[Markup, str]:
    interface_language_code = cast(str, flask.g.interface_language_code)
    language_codes = [interface_language_code, *lang_fallbacks(interface_language_code), 'en']
    translations = flask.current_app.extensions['toolforge_i18n'].translations
    for language_code in language_codes:
        try:
            text = translations[language_code][message_code]
        except LookupError:
            continue
        else:
            return Markup(text), language_code
    raise ValueError(f'Message {message_code} not found in {language_codes}')


def _message_qqx(message_code: str, **kwargs: object) -> Markup:
    message = Markup('(')
    message += message_code
    if kwargs:
        message += ': '
        first = True
        for key, value in kwargs.items():
            if first:
                first = False
            else:
                message += ', '
            message += key
            message += '='
            message += repr(value)
    message += ')'
    return message


def message(message_code: str, **kwargs: object) -> Markup:
    """Format an interface message in the user interface language.

    The kwargs may contain (named) arguments,
    using the argument names defined in TranslationsConfig.varaibles.

    This method is available as a template global, and is usually used there
    (but may also be imported and called from Python code).
    """
    if flask.g.qqx:
        return _message_qqx(message_code, **kwargs)
    message, language = _message_with_language(message_code)
    if kwargs:
        config = flask.current_app.extensions['toolforge_i18n'].config
        formatter = I18nFormatter(
            locale_identifier=config.language_code_to_babel(language),
            get_gender=config.get_gender,
        )
        # I18nFormatter returns Markup given Markup
        message = cast(Markup, formatter.format(message, **kwargs))
    return add_lang_if_needed(message, language)


class ToolforgeI18n:
    """Flask extension for toolforge_i18n.

    Basic usage:

        app = flask.Flask(__name__)
        i18n = ToolforgeI18n(app)
    """

    def __init__(
        self,
        app: flask.Flask | None = None,
        interface_language_code: Callable[[dict[str, dict[str, str]]], str] = interface_language_code_from_request,
    ):
        import tool_translations_config

        self.config = tool_translations_config.config
        self.translations, self.documentation = load_translations(self.config)
        self.interface_language_code = interface_language_code
        if app is not None:
            self.init_app(app)

    def init_app(self, app: flask.Flask) -> None:
        app.extensions['toolforge_i18n'] = self
        app.add_template_global(message)
        app.add_template_filter(lang_mw_to_bcp47)
        app.add_template_global(push_html_lang)
        app.add_template_global(pop_html_lang)
        app.before_request(init_html_language_codes)
        app.after_request(assert_html_language_codes_empty)

        @app.before_request
        def init_interface_language_code() -> None:
            interface_language_code = self.interface_language_code(self.translations)
            if interface_language_code == 'qqx':
                flask.g.interface_language_code = 'en'
                flask.g.qqx = True
            else:
                flask.g.interface_language_code = interface_language_code
                flask.g.qqx = False



import warnings
from collections.abc import Callable
from typing import Any, Never

import bs4
from markupsafe import Markup

import toolforge_i18n._formatters as formatters
from toolforge_i18n._translations import TranslationsConfig


def unused(*_args: object, **_kwargs: object) -> Never:
    raise RuntimeError('This function should not be called!')


def check_message_keys(translations: dict[str, dict[str, str]], language_code: str) -> None:
    """Check for extraneous message keys.

    The translations for the given language should not contain
    any message keys that are not in the English translations.
    """
    language_keys = set(translations[language_code].keys())
    english_keys = set(translations['en'].keys())
    extra_keys = language_keys.difference(english_keys)
    assert not extra_keys


def check_message_documentation(
    translations_config: TranslationsConfig, translations: dict[str, dict[str, str]], documentation: dict[str, str]
) -> None:
    """Check message documentation keys.

    All English messages should have a documentation message,
    and all documentation messages should match an existing English message.
    """
    documented_keys = set(documentation.keys())
    expected_documented_keys = {key for key in translations['en'] if key not in translations_config.derived_messages}
    assert documented_keys == expected_documented_keys


def check_message_html_elements(
    translations_config: TranslationsConfig,
    translations: dict[str, dict[str, str]],
    language_code: str,
    message_key: str,
) -> None:
    """Check HTML elements in the message translation.

    Parses the message as HTML,
    and checks that all the element and attribute names are allowed by the config.

    This test protects against malicious translations.
    If it fails for an English source message after you changed it,
    you probably need to update the tool translations config to allow the element or attribute name.
    If it fails for a translation in another language, you need to be very careful.
    If you are sure that the translation is safe,
    you may want to update the tool translations config in this case as well;
    but if the translation contains HTML that should not be allowed
    (e.g. "script", "style" or "img" elements or any attributes starting with "on"),
    you should not use it and instead revert it on translatewiki.net.
    """
    message = translations[language_code][message_key]
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=bs4.MarkupResemblesLocatorWarning)
        soup = bs4.BeautifulSoup(message, features='html.parser')
    for element in soup.find_all():
        assert element.name in translations_config.allowed_html_elements
        allowed_attributes = (
            translations_config.allowed_html_elements[element.name] | translations_config.allowed_global_attributes
        )
        for attr in element.attrs:
            assert attr in allowed_attributes


def check_message_variables(
    translations_config: TranslationsConfig,
    translations: dict[str, dict[str, str]],
    language_code: str,
    message_key: str,
) -> None:
    """Test that the translation uses variables correctly.

    This test checks that all the variables are used in the translation
    (except that PLURAL and GENDER variables are allowed to be unused if not needed),
    with the right format, and never used with the wrong format.
    The translation isn’t actually formatted in this test.

    See the TranslationConfig.variables docstring
    for the meaning of the different variable names / prefixes.
    """
    message = translations[language_code][message_key]
    for variable in translations_config.variables.get(message_key, []):
        if variable == 'url' or variable.startswith('url_'):
            assert '{' + variable + '!h:' in message
            assert '{' + variable + '!g:' not in message
            assert '{' + variable + '!p:' not in message
            assert '{' + variable + '!l}' not in message
            assert '{' + variable + '}' not in message
        elif variable == 'user_name' or variable.startswith('user_name_'):
            assert '{' + variable + '!g:' in message or '{' + variable not in message
            assert '{' + variable + '!h:' not in message
            assert '{' + variable + '!p:' not in message
            assert '{' + variable + '!l}' not in message
            assert '{' + variable + '}' not in message
        elif variable == 'num' or variable.startswith('num_'):
            assert '{' + variable + '!p:' in message or '{' + variable not in message
            assert '{' + variable + '!h:' not in message
            assert '{' + variable + '!g:' not in message
            assert '{' + variable + '!l}' not in message
            # assert '{' + variable + '}' not in message  # allowed, e.g. {{PLURAL:$1||$1 forms}}
        elif variable == 'list' or variable.startswith('list_'):
            assert '{' + variable + '!l}' in message
            assert '{' + variable + '!h:' not in message
            assert '{' + variable + '!g:' not in message
            assert '{' + variable + '!p:' not in message
            assert '{' + variable + '}' not in message
        else:
            assert '{' + variable + '}' in message
            assert '{' + variable + '!h:' not in message
            assert '{' + variable + '!g:' not in message
            assert '{' + variable + '!p:' not in message
            assert '{' + variable + '!l}' not in message


def check_message_formatting(
    translations_config: TranslationsConfig,
    translations: dict[str, dict[str, str]],
    language_code: str,
    message_key: str,
) -> None:
    """Test that the translation uses variables correctly.

    This test actually formats the translation,
    with example values for different variable formats,
    and makes sure that nothing crashes and the formatted variables appear in the result.

    One of the problems this catches is using a URL variable inside a GENDER variable,
    e.g. in "{{GENDER:$2|Logged in as $1.}}" –
    this does not work because, when the URL in $1 is expanded first,
    a colon inside it (e.g. in "https://...") is misinterpreted as the ":" separator in GENDER.
    Such translations need to be changed,
    e.g. to "{{GENDER:$2|Logged in}} as $1."

    See the TranslationConfig.variables docstring
    for the meaning of the different variable names / prefixes.
    """
    message = translations[language_code][message_key]

    url = 'https://example.com/test?foo=bar#baz'
    assert_contains = []
    get_gender: Callable[[str], str] = unused
    params: dict[str, Any] = {}
    for variable in translations_config.variables.get(message_key, []):
        if variable == 'url' or variable.startswith('url_'):
            params[variable] = url
            assert_contains.append(url)
        elif variable == 'user_name' or variable.startswith('user_name_'):
            get_gender = lambda user_name: 'n'  # noqa: E731, ARG005
            params[variable] = 'User name'
        elif variable == 'num' or variable.startswith('num_'):
            params[variable] = 123
        elif variable == 'list' or variable.startswith('list_'):
            item_1 = Markup('<a href="{url}">{variable} A</a>').format(url=url, variable=variable)
            item_2 = Markup('<a href="{url}">{variable} B</a>').format(url=url, variable=variable)
            params[variable] = [item_1, item_2]
            assert_contains.append(item_1)
            assert_contains.append(item_2)
        else:
            value = Markup('<a href="{url}">{variable}</a>').format(url=url, variable=variable)
            params[variable] = value
            assert_contains.append(value)

    formatter = formatters.I18nFormatter(
        locale_identifier=translations_config.language_code_to_babel(language_code),
        get_gender=get_gender,
    )
    formatted = formatter.format(message, **params)

    for assertion in assert_contains:
        assert assertion in formatted


def check_all_translations(
    translations_config: TranslationsConfig,
    translations: dict[str, dict[str, str]],
    documentation: dict[str, str],
) -> None:
    if not __debug__:
        raise AssertionError(
            'Translation checks are implemented using `assert`; you must run this code without optimization (-O) enabled!'
        )
    check_message_documentation(translations_config, translations, documentation)
    for language_code in translations:
        check_message_keys(translations, language_code)
        for message_key in translations[language_code]:
            check_message_html_elements(translations_config, translations, language_code, message_key)
            check_message_variables(translations_config, translations, language_code, message_key)
            check_message_formatting(translations_config, translations, language_code, message_key)



try:
    import pytest
except ModuleNotFoundError:
    pass
else:
    pytest.register_assert_rewrite('toolforge_i18n._translations_checks')


from toolforge_i18n._flask_things import (
    ToolforgeI18n,
    add_lang_if_needed,
    interface_language_code_from_request,
    message,
    pop_html_lang,
    push_html_lang,
)
from toolforge_i18n._formatters import I18nFormatter
from toolforge_i18n._get_gender import get_gender_by_user_name
from toolforge_i18n._language_info import lang_autonym, lang_bcp47_to_mw, lang_dir, lang_fallbacks, lang_mw_to_bcp47
from toolforge_i18n._translations import TranslationsConfig, language_code_to_babel, load_translations
from toolforge_i18n._user_agent import get_user_agent, set_user_agent



import dataclasses
import os
from collections.abc import Iterable
from pathlib import Path
from typing import Any, Optional, cast

import pytest

from toolforge_i18n._translations import TranslationsConfig, load_translations
from toolforge_i18n._translations_checks import (
    check_message_documentation,
    check_message_formatting,
    check_message_html_elements,
    check_message_keys,
    check_message_variables,
)

translations_config_key = pytest.StashKey[TranslationsConfig]()
translations_key = pytest.StashKey[dict[str, dict[str, str]]]()
documentation_key = pytest.StashKey[dict[str, str]]()
language_code_key = pytest.StashKey[dict[str, str]]()
message_key_key = pytest.StashKey[str]()


def _in_toolforge_i18n(session: pytest.Session) -> bool:
    """Determine whether we’re in toolforge_i18n itself.

    In this case, the plugin should essentially be skipped:
    there’s no tool_translations_config here.
    """
    return session.path.name == 'toolforge_i18n'


def pytest_collection(session: pytest.Session) -> None:
    """Initialize the plugin by loading the translations.

    Tools configure their translations by creating an instance of TranslationsConfig
    and exporting it as the config member of a tool_translations_config module.
    Here, we import the config from there, load the translations and message documentation,
    and stash it all in the pytest session.

    If as a tool developer you get an ImportError from this function,
    you might have forgotten to create a tool_translations_config.py file;
    please see the toolforge_i18n README for instructions.
    """
    if _in_toolforge_i18n(session):
        return

    try:
        import tool_translations_config
    except ModuleNotFoundError:
        # pytest may not have added the session path to the sys.path yet;
        # try to do it now, respecting --import-mode if possible
        # (but with --import-mode=importlib, just add to sys.path anyway,
        # because I don’t know how to use importlib) –
        # improvement suggestions welcome :')
        import sys

        import_mode = session.config.getoption('--import-mode')
        if import_mode == 'append':
            sys.path.append(str(session.path))
        else:
            sys.path.insert(0, str(session.path))
        import tool_translations_config

    translations_config = tool_translations_config.config
    translations, documentation = load_translations(dataclasses.replace(translations_config, check_translations=False))
    session.stash[translations_config_key] = translations_config
    session.stash[translations_key] = translations
    session.stash[documentation_key] = documentation


def pytest_collect_directory(parent: pytest.Directory, path: Path) -> Optional['I18nDirectory']:
    """Custom collection for the i18n directory containing message JSON files."""
    if _in_toolforge_i18n(parent.session):
        return None
    translations_config = parent.session.stash[translations_config_key]
    if path == parent.session.path / translations_config.directory:
        return I18nDirectory.from_parent(parent, path=path)
    return None


class I18nDirectory(pytest.Directory):
    """Custom collector for the i18n directory.

    Collects qqq.json as DocumentationFile,
    and all other JSON files as TranslationsFile.
    Other tests in this directory cannot be collected;
    the collector tries to detect if users put tests there anyway,
    and will warn on test_*.py files or any subdirectories.
    """

    def collect(self) -> Iterable[pytest.Item | pytest.Collector]:
        files: dict[str, pytest.Item | pytest.Collector] = {}
        for direntry in os.scandir(self.path):
            if not direntry.is_file():
                if direntry.is_dir():
                    self.warn(
                        UserWarning(
                            f'Unexpected subdirectory in {self.path}; tests will not be collected here: {direntry.name}'
                        )
                    )
                continue
            path = Path(direntry.path)
            if path.suffix != '.json':
                if path.suffix == '.py' and path.stem.startswith('test_'):
                    self.warn(
                        UserWarning(
                            f'Unexpected test file in {self.path}; tests will not be collected here: {direntry.name}'
                        )
                    )
                continue
            if path.stem == 'qqq':
                files[path.stem] = DocumentationFile.from_parent(self, path=path)
            else:
                files[path.stem] = TranslationsFile.from_parent(self, path=path, language_code=path.stem)
        for stem in sorted(files):
            yield files[stem]


class TranslationsFile(pytest.File):
    """Custom collector for a message JSON file.

    Yields various Functions for the different checks from translations_checks.
    The language code and (where applicable) message key are added to the Function’s stash,
    from where the language_code and mesage_key fixtures (below) later retrieve it.
    """

    def __init__(self, *, path: Path, language_code: str, **kwargs: Any):
        super().__init__(
            path=path,
            **kwargs,
        )
        self.language_code = language_code

    def collect(self) -> Iterable[pytest.Item | pytest.Collector]:
        if self.language_code != 'en':
            fun = pytest.Function.from_parent(
                self,
                name='check_message_keys',
                callobj=check_message_keys,
            )
            fun.stash[language_code_key] = self.language_code  # type: ignore[misc]
            yield fun
        translations = self.session.stash[translations_key]
        for message_key in translations[self.language_code]:
            fun = pytest.Function.from_parent(
                self,
                name=f'check_message_html_elements[{message_key}]',
                callobj=check_message_html_elements,
            )
            fun.stash[language_code_key] = self.language_code  # type: ignore[misc]
            fun.stash[message_key_key] = message_key
            yield fun
            fun = pytest.Function.from_parent(
                self,
                name=f'check_message_variables[{message_key}]',
                callobj=check_message_variables,
            )
            fun.stash[language_code_key] = self.language_code  # type: ignore[misc]
            fun.stash[message_key_key] = message_key
            yield fun
            fun = pytest.Function.from_parent(
                self,
                name=f'check_message_formatting[{message_key}]',
                callobj=check_message_formatting,
            )
            fun.stash[language_code_key] = self.language_code  # type: ignore[misc]
            fun.stash[message_key_key] = message_key
            yield fun


class DocumentationFile(pytest.File):
    """Custom collector for a message documentation JSON file (qqq.json).

    Yields one Function for check_message_documentation.
    """

    def __init__(self, *, path: Path, **kwargs: Any):
        super().__init__(
            path=path,
            **kwargs,
        )

    def collect(self) -> Iterable[pytest.Item | pytest.Collector]:
        yield pytest.Function.from_parent(
            self,
            name='check_message_documentation',
            callobj=check_message_documentation,
        )


@pytest.fixture(scope='session')
def translations_config(request: pytest.FixtureRequest) -> TranslationsConfig:
    """Fixture for the tool translations config."""
    return request.session.stash[translations_config_key]


@pytest.fixture(scope='session')
def translations(request: pytest.FixtureRequest) -> dict[str, dict[str, str]]:
    """Fixture for the tool’s translations."""
    return request.session.stash[translations_key]


@pytest.fixture(scope='session')
def documentation(request: pytest.FixtureRequest) -> dict[str, str]:
    """Fixture for the tool’s documentation messages."""
    return request.session.stash[documentation_key]


@pytest.fixture
def language_code(request: pytest.FixtureRequest) -> str:
    """Fixture for a translation language code.

    Tests can use this fixture,
    together with the translations_config and (optionally) message_key fixtures,
    to implement additional checks for translations.
    The fixture is parameterized and yields all language codes for which translations exist
    (including 'en', but excluding 'qqq').
    """
    # Implementation note: normal tests actually get this fixture from pytest_generate_tests() below;
    # this fixture is only called by the tests added by this plugin,
    # which are added by the custom collectors above and bypass the Metafunc mechanism.
    return cast(str, request.node.stash[language_code_key])


@pytest.fixture
def message_key(request: pytest.FixtureRequest) -> str:
    """Fixture for a translation message key.

    Tests can use this fixture,
    together with the translations_config and language_code fixtures,
    to implement additional checks for translations.
    The fixture is parameterized and yields all message keys defined in the translations;
    if a test uses both language_code and message_key,
    then the test will only be called with combinations of both that exist in the translations
    (i.e. for each language_code, the message_key will skip messages
    that have not been translated into this language code yet).
    """
    # Implementation note: normal tests actually get this fixture from pytest_generate_tests() below;
    # this fixture is only called by the tests added by this plugin,
    # which are added by the custom collectors above and bypass the Metafunc mechanism.
    return cast(str, request.node.stash[message_key_key])


# https://docs.pytest.org/en/latest/how-to/parametrize.html#basic-pytest-generate-tests-example
# the below is not used for our own functions, but supports users’ own additional tests
def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    """Parameterize tests that use language_code and/or message_key.

    This implements the parametrization of the language_code and message_key fixtures,
    as documented above,
    effectively replacing the fixtures themselves for all normal tests
    (after we’ve parametrized the Metafunc here,
    pytest won’t have to call the actual fixture anymore).
    """
    if _in_toolforge_i18n(metafunc.definition.session):
        return
    if 'message_key' in metafunc.fixturenames:
        translations = metafunc.definition.session.stash[translations_key]
        if 'language_code' in metafunc.fixturenames:
            metafunc.parametrize(
                ('language_code', 'message_key'),
                [
                    (language_code, message_key)
                    for language_code in translations
                    for message_key in translations[language_code]
                ],
            )
        else:
            metafunc.parametrize('message_key', translations['en'].keys())
    elif 'language_code' in metafunc.fixturenames:
        translations = metafunc.definition.session.stash[translations_key]
        metafunc.parametrize('language_code', translations.keys())



from typing import Any

import pytest
from flask import Flask
from markupsafe import Markup

import toolforge_i18n._user_agent
from toolforge_i18n._flask_things import (
    add_lang_if_needed,
    assert_html_language_codes_empty,
    init_html_language_codes,
    interface_language_code_from_request,
    pop_html_lang,
    push_html_lang,
)

toolforge_i18n._user_agent.set_user_agent(  # noqa: SLF001
    'toolforge-i18n test (https://gitlab.wikimedia.org/lucaswerkmeister/toolforge_i18n/; mail@lucaswerkmeister.de)'
)


app = Flask(__name__)


def test_html_lang_stack() -> None:
    with app.test_request_context():
        init_html_language_codes()
        assert push_html_lang('en') == Markup('lang="en" dir="ltr"')
        assert push_html_lang('ar') == Markup('lang="ar" dir="rtl"')
        assert pop_html_lang('ar') == Markup('')
        assert pop_html_lang('en') == Markup('')
        response: Any = 'unused argument that should be returned unchanged'
        assert assert_html_language_codes_empty(response) == response


def test_html_lang_stack_wrong_order() -> None:
    with app.test_request_context():
        init_html_language_codes()
        assert push_html_lang('en') == Markup('lang="en" dir="ltr"')
        assert push_html_lang('ar') == Markup('lang="ar" dir="rtl"')
        with pytest.raises(AssertionError):
            pop_html_lang('en')


def test_html_lang_stack_not_empty() -> None:
    with app.test_request_context():
        init_html_language_codes()
        assert push_html_lang('en') == Markup('lang="en" dir="ltr"')
        assert push_html_lang('ar') == Markup('lang="ar" dir="rtl"')
        assert pop_html_lang('ar') == Markup('')
        response: Any = 'unused argument that should be returned unchanged'
        with pytest.raises(AssertionError):
            assert_html_language_codes_empty(response)


def test_add_lang_if_needed() -> None:
    with app.test_request_context():
        init_html_language_codes()
        push_html_lang('en')
        assert add_lang_if_needed(Markup('msg'), 'ar') == Markup('<span lang="ar" dir="rtl">msg</span>')


def test_add_lang_if_needed_unneeded() -> None:
    with app.test_request_context():
        init_html_language_codes()
        push_html_lang('en')
        push_html_lang('ar')
        assert add_lang_if_needed(Markup('msg'), 'ar') == Markup('msg')


def test_interface_language_code_from_request_params() -> None:
    with app.test_request_context('/?uselang=simple'):
        translations: dict[str, dict[str, str]] = {
            'en': {},
            'en-us': {},
            'simple': {},
        }
        assert interface_language_code_from_request(translations) == 'simple'


def test_interface_language_code_from_request_headers() -> None:
    with app.test_request_context(headers=[('Accept-Language', 'de;q=0.9, en-simple;q=0.8, en;q=0.7')]):
        translations: dict[str, dict[str, str]] = {
            'en': {},
            'en-us': {},
            'simple': {},
        }
        assert interface_language_code_from_request(translations) == 'simple'


def test_interface_language_code_from_request_nothing() -> None:
    with app.test_request_context():
        translations: dict[str, dict[str, str]] = {
            'en': {},
            'en-us': {},
            'simple': {},
        }
        assert interface_language_code_from_request(translations) == 'en'
